@@ -23,7 +23,7 @@ class Checkpoint:
 @dataclass
 class ComparisonTask:
     batch_id: int
-    source_key_vals: list[dict]
+    source_rows: list[dict]
     different: list[dict]
 
 
@@ -88,6 +88,16 @@ def get_elapsed_time(st: float, ndigits=None) -> int | float:
     return round(time.time() - st, ndigits)
 
 
+def extract_keyvals(row: dict, keys: list[tuple[str, str]]):
+    _keys = [item[0] for item in keys]
+    new_dict = {}
+    for key in row:
+        if key in _keys:
+            new_dict[key] = row[key]
+
+    return new_dict
+
+
 class MysqlTableCompare:
     def __init__(
         self,
@@ -134,9 +144,9 @@ class MysqlTableCompare:
         cts.sort(key=lambda task: task.batch_id)
 
         while cts and cts[0].batch_id == batch_id:
-            _, source_key_vals, diff_rows = astuple(cts.pop(0))
+            _, source_rows, diff_rows = astuple(cts.pop(0))
 
-            self.processed_rows_number += len(source_key_vals)
+            self.processed_rows_number += len(source_rows)
 
             if diff_rows:
                 self.write_different(diff_rows)
@@ -145,7 +155,7 @@ class MysqlTableCompare:
             if self.processed_rows_number % 10000 == 0 or diff_rows:
                 self.checkpoint.processed = self.processed_rows_number
                 self.checkpoint.different = self.different_rows_number
-                self.checkpoint.row = source_key_vals[-1]
+                self.checkpoint.row = extract_keyvals(source_rows[-1], self.source_table_keys)
                 self.write_checkpoint()
                 self.logger.info(f"checkpoint: {self.checkpoint}")
 
@@ -194,28 +204,28 @@ class MysqlTableCompare:
                 cur.execute(_stmt, tuple(params))
                 return cur.fetchall()
 
-    def get_full_table_keys_order(self, keycols: list[tuple[str, str]], ckpt_row: dict = None):
+    def get_full_table_keys_order(self, ckpt_row: dict = None):
         _keyval = ckpt_row
         # select * from where 1 = 1 and ((a > xxx) or (a = xxx and b > yyy) or (a = xxx and b = yyy and c > zzz)) order by a,b,c limit checksize
-        _key_colns = ", ".join([f"`{col[0]}`" for col in keycols])
+        _key_colns = ", ".join([f"`{col[0]}`" for col in self.source_table_keys])
 
-        for _, column_type in keycols:
+        for _, column_type in self.source_table_keys:
             if column_type in ["int", "double", "char", "date", "decimal", "varchar", "bigint", "tinyint", "smallint"]:
                 pass
             else:
                 raise ValueError(f"Data type: [{column_type}] is not supported yet.")
 
         where_conditions = []
-        for end_idx in range(len(keycols)):
+        for end_idx in range(len(self.source_table_keys)):
             condition_parts = []
-            for i, (column_name, _) in enumerate(keycols[: end_idx + 1]):
+            for i, (column_name, _) in enumerate(self.source_table_keys[: end_idx + 1]):
                 operator = ">" if i == end_idx else "="
                 condition_parts.append(f"`{column_name}` {operator} %s")
             where_conditions.append(" and ".join(condition_parts))
         where_clause = "WHERE" + "(" + ") or (".join(where_conditions) + ")"
 
-        statement_with_condition = f"SELECT {_key_colns} FROM {self.src_database}.{self.src_table} {where_clause} ORDER BY {_key_colns} LIMIT {self.limit_size}"
-        statement_without_condition = f"SELECT {_key_colns} FROM {self.src_database}.{self.src_table} ORDER BY {_key_colns} LIMIT {self.limit_size}"
+        statement_with_condition = f"SELECT * FROM {self.src_database}.{self.src_table} {where_clause} ORDER BY {_key_colns} LIMIT {self.limit_size}"
+        statement_without_condition = f"SELECT * FROM {self.src_database}.{self.src_table} ORDER BY {_key_colns} LIMIT {self.limit_size}"
 
         with connect(**self.source_dsn) as source_con:
             source_con: MySQLConnection
@@ -227,8 +237,8 @@ class MysqlTableCompare:
                 _rows_queue = deque()
 
                 if _keyval:
-                    for end_idx in range(len(keycols)):
-                        for i, (column_name, _) in enumerate(keycols[: end_idx + 1]):
+                    for end_idx in range(len(self.source_table_keys)):
+                        for i, (column_name, _) in enumerate(self.source_table_keys[: end_idx + 1]):
                             _params.append(_keyval[column_name])
 
                 statement = statement_with_condition if _params else statement_without_condition
@@ -247,39 +257,30 @@ class MysqlTableCompare:
                     self.logger.debug(f"full table query - fetch once query: {get_elapsed_time(_start_time_1, 2)}s, fetch rows count: {_rows_cnt}.")
 
                     while _rows_queue:
-                        _start_time_2 = time.time()
                         rows = []
                         for _ in range(min(self.fetch_size, len(_rows_queue))):
                             rows.append(_rows_queue.popleft())
 
-                        _keyval = rows[-1]
+                        last_row = rows[-1]
+                        _keyval = extract_keyvals(last_row, self.source_table_keys)
                         yield rows
-                        self.logger.debug(f"full table query - yielded rows: {get_elapsed_time(_start_time_2, 2)}s,.")
                 else:
                     self.logger.debug(f"full table query - fetch all rows: {get_elapsed_time(_start_time_0, 2)}s.")
                     return
 
     def compare_rows_by_keys(self, ct: ComparisonTask):
-        _task_configs = {
-            "source": (self.source_conpool, self.src_database, self.src_table, ct.source_key_vals),
-            "target": (self.target_conpool, self.dst_database, self.dst_table, ct.source_key_vals),
-        }
-
-        _tasks_results = {"source": [], "target": []}
-
         _start_time = time.time()
 
-        with concurrent.futures.ThreadPoolExecutor(max_workers=2) as executor:
-            futures = {executor.submit(self.get_table_rows_by_keys, *args): name for name, args in _task_configs.items()}
-            try:
-                _tasks_results = {name: future.result() for future, name in futures.items()}
-            except Exception as e:
-                raise Exception(f"compare rows failure - batch[1] - error: {e}.")
+        keyval_list = []
+        for row in ct.source_rows:
+            keyval_list.append(extract_keyvals(row, self.source_table_keys))
 
-        ct.different += list(itertools.filterfalse(lambda x: x in _tasks_results["target"], _tasks_results["source"]))
+        target_rows = self.get_table_rows_by_keys(self.target_conpool, self.dst_database, self.dst_table, keyval_list)
+
+        ct.different += list(itertools.filterfalse(lambda x: x in target_rows, ct.source_rows))
 
         self.logger.debug(
-            f"compare rows success - batch[{ct.batch_id}] - elapsed time {get_elapsed_time(_start_time,2)}s, rows: {len(_tasks_results['source'])}, different: {len(ct.different)}."
+            f"compare rows success - batch[{ct.batch_id}] - elapsed time {get_elapsed_time(_start_time,2)}s, rows: {len(ct.source_rows)}, different: {len(ct.different)}."
         )
 
         return ct
@@ -289,15 +290,15 @@ class MysqlTableCompare:
         # batch_id, source_key_vals, diff_rows = _tasks.pop(0)
         _comparison_tasks_container: list[ComparisonTask] = []
 
-        fulltasks = enumerate(self.get_full_table_keys_order(self.source_table_keys, checkpoint_row), start=1)
+        fulltasks = enumerate(self.get_full_table_keys_order(checkpoint_row), start=1)
 
         with concurrent.futures.ThreadPoolExecutor(max_workers=self.parallel) as executor:
 
             def generator_futures(n):
                 try:
                     return [
-                        executor.submit(self.compare_rows_by_keys, ComparisonTask(batch_id, source_key_vals, []))
-                        for batch_id, source_key_vals in itertools.islice(fulltasks, n)
+                        executor.submit(self.compare_rows_by_keys, ComparisonTask(batch_id, source_row, []))
+                        for batch_id, source_row in itertools.islice(fulltasks, n)
                     ]
                 except Exception as e:
                     raise Exception(f"full table query failure - error: {e}.")
